import logging
import numpy as np
import re

from copy import deepcopy
from random import randint

from bloc.util import dumpJsonToFile
from bloc.util import genericErrorInfo
from bloc.util import gen_bloc_variant_tf_mat
from bloc.util import get_tf_matrix
#from bloc.util import merge_bloc_matrices
from bloc.util import update_bloc_model

from bloc.MarkovChain import BLOCMarkovChain

from sklearn.metrics import pairwise_distances

logger = logging.getLogger('bloc.bloc')

def extract_bloc_from_segments(testing_segment, training_segments):

    experiment = []
    for segment in training_segments:
        experiment.append({'testing_seg': testing_segment, 'training_seg': segment})

    return experiment

def generate_bloc_expr_template(all_screenames_bloc):

    max_segment_number = -1
    min_segment_number = 1000000000
    
    #get max and min segment_number
    for user in all_screenames_bloc:

        segment_numbers = user['bloc_segments']['segments'].keys()
        loc_max = max( segment_numbers )
        loc_min = min( segment_numbers )

        if( loc_max > max_segment_number ):
            max_segment_number = loc_max

        if( loc_min < min_segment_number ):
            min_segment_number = loc_min
    

    experiment_buckets = {}
    src_target_experiment_buckets = []
    for src_segment_number in range(max_segment_number, min_segment_number, -1):
    
        experiment_buckets[src_segment_number] = []
        for tgt_segment_number in range(src_segment_number-1, min_segment_number-1, -1):
            experiment_buckets[src_segment_number].append( list(range(src_segment_number-1, tgt_segment_number-1, -1)) )
        
        src_target_experiment_buckets.append(
            {'testing_segment': src_segment_number, 
            'training_segments': experiment_buckets[src_segment_number]
        })


    #extract bloc from src_segment and combined(single_target) for all users
    all_experiments = []
    for expr_run in src_target_experiment_buckets:
        all_experiments += extract_bloc_from_segments( expr_run['testing_segment'], expr_run['training_segments'] )


    flat_experiment = {}
    for expr_run in all_experiments:
        
        s_seg = expr_run['testing_seg']
        flat_experiment.setdefault(s_seg, [])
        flat_experiment[s_seg].append( expr_run['training_seg'] )


    semantic_exper = []
    for test_seg, train_seg in flat_experiment.items():
        semantic_exper.append({'testing_segment': test_seg, 'training_segments': train_seg})
    
    return semantic_exper

def obsolete_get_user_bloc_for_dim(bloc_segments, segment_numbers, seg_dimensions ):

    doc_lst = []
    for seg_num in segment_numbers:

        if( seg_num not in bloc_segments ):
            continue

        seg_dimensions_docs = ''
        for seg_dim in seg_dimensions.split(', '):
        
            seg_dim = seg_dim.strip()
            if( seg_dim not in bloc_segments[seg_num] ):
                continue

            seg_dimensions_docs = seg_dimensions_docs + ' * ' + bloc_segments[seg_num][seg_dim]

        if( seg_dimensions_docs.startswith(' * ') ):
            seg_dimensions_docs = seg_dimensions_docs[3:]

        if( seg_dimensions_docs.endswith(' * ') ):
            seg_dimensions_docs = seg_dimensions_docs[:-3]

        doc_lst.append( seg_dimensions_docs )

    return doc_lst

def get_user_bloc_for_dim(bloc_segments, segment_numbers, seg_dimensions ):

    doc_lst = []
    for seg_num in segment_numbers:

        if( seg_num not in bloc_segments ):
            continue

        seg_dimensions_docs = {}
        for seg_dim in seg_dimensions.split(', '):
        
            seg_dim = seg_dim.strip()
            if( seg_dim not in bloc_segments[seg_num] ):
                continue

            seg_dimensions_docs[seg_dim] = bloc_segments[seg_num][seg_dim]

        doc_lst.append( seg_dimensions_docs )

    return doc_lst

def combine_text_across_segs_bloc_dims(seg_dim_text):

    doc_lst = []

    for segment in seg_dim_text:

        seg_dimensions_docs = ''
        for dim, bloc_seq in segment.items():
            seg_dimensions_docs = seg_dimensions_docs + ' * ' + bloc_seq

        if( seg_dimensions_docs.startswith(' * ') ):
            seg_dimensions_docs = seg_dimensions_docs[3:]

        if( seg_dimensions_docs.endswith(' * ') ):
            seg_dimensions_docs = seg_dimensions_docs[:-3]

        doc_lst.append( seg_dimensions_docs )

    doc_lst = ' ** '.join(doc_lst)
    return doc_lst

def rm_absent_segment(cur_segments, user_segments):

    dedup_set = set()
    final_segments = []
    
    for segment in cur_segments:
        
        new_segment = []
        for seg in segment:
            if( seg in user_segments ):
                new_segment.append( seg )

        ky = str(new_segment)
        if( ky in dedup_set ):
            continue

        dedup_set.add(ky)
        
        if( len(new_segment) != 0 ):
            final_segments.append( new_segment )

    return final_segments

def gen_expr_doc_lst_for_single_user(user, all_experiments, expr_bloc_dims):

    logger.debug('\ngen_expr_doc_lst_for_single_user(), user: ' + user['screen_name'])

    for expr_run in all_experiments:

        if( 'testing_segment' not in expr_run or 'training_segments' not in expr_run ):
            continue
        
        test_seg = expr_run['testing_segment']
        test_doc_lst = get_user_bloc_for_dim( user['bloc_segments']['segments'], [test_seg], expr_bloc_dims )

        if( len(test_doc_lst) == 0 ):
            #no test docs so no need for training docs
            continue
        
        logger.debug( 'experiment: ' + str(expr_run['testing_segment']) + ' vs ' + str(expr_run['training_segments']) )
        expr_run['training_segments'] = rm_absent_segment( expr_run['training_segments'], user['bloc_segments']['segments'] )
        logger.debug( 'experiment: ' + str(expr_run['testing_segment']) + ' vs ' + str(expr_run['training_segments']) + ' (actually)')
        
        if( len(expr_run['training_segments']) != 0 ):
            
            expr_run['training_doc_lst'] = []
            #don't add test if training is absent
            #expr_run['testing_doc_lst'] = { 'text': ' ** '.join(test_doc_lst), 'segments': [test_seg]}
            expr_run['testing_doc_lst'] = { 'text': test_doc_lst, 'segments': [test_seg]}
            
        for train in expr_run['training_segments']:
            
            train_doc_lst = get_user_bloc_for_dim( user['bloc_segments']['segments'], train, expr_bloc_dims )
            expr_run['training_doc_lst'].append({
                #'text': ' ** '.join(train_doc_lst),
                'text': train_doc_lst,
                'segments': train
            })
            
            
            logger.debug('\t' + str(test_seg) + ' vs ' + str(train))
            logger.debug('\t\ttest_doc_lst: ' + str(test_doc_lst))
            logger.debug('\t\ttrain_doc_lst: ' + str(train_doc_lst))
    
    logger.debug('')
    for expr_run in all_experiments:
        
        #remove testing_segment and training_segments
        if( 'testing_segment' in expr_run ):
            del expr_run['testing_segment']

        if(  'training_segments' in expr_run ):
            del expr_run['training_segments']

def gen_expr_doc_lst_for_users( all_screenames_bloc, all_experiments, expr_bloc_dims ):

    for i in range( len(all_screenames_bloc) ):

        user = all_screenames_bloc[i]
        user['bloc_experiment_docs'] = deepcopy(all_experiments)
        user['bloc_experiment_docs_dim'] = expr_bloc_dims

        gen_expr_doc_lst_for_single_user( user, user['bloc_experiment_docs'], expr_bloc_dims )

        #rm empty experiment docs
        user['bloc_experiment_docs'] = [ doc for doc in user['bloc_experiment_docs'] if len(doc) != 0 ]

def gen_ngram_tf_matrix(all_screenames_bloc, ngram, token_pattern, matrix_key='tf_matrix_normalized', tf_matrix_norm='l1', tf_idf_norm='l2' ):

    logger.debug('\ngen_ngram_tf_matrix()')
    tf_matrix = []

    for i in range( len(all_screenames_bloc) ):
        user = all_screenames_bloc[i]

        if( 'bloc_experiment_docs' not in user ):
            continue

        logger.debug('user: ' + user['screen_name'])

        for j in range( len(user['bloc_experiment_docs']) ):
            
            expr_doc = user['bloc_experiment_docs'][j]
            if( 'testing_doc_lst' not in expr_doc and 'training_doc_lst' not in expr_doc ):
                continue

            logger.debug('\ttesting doc: ' + str(expr_doc['testing_doc_lst']))

            tf_matrix.append({
                'screen_name': user['screen_name'],
                'segments': expr_doc['testing_doc_lst']['segments'],
                'text': combine_text_across_segs_bloc_dims( expr_doc['testing_doc_lst']['text'] ),
                'text_structured': expr_doc['testing_doc_lst']['text'],
                'id': str(i) + ', ' + str(j),
                'user_indx': i,
                'experiment_indx': j
            })

            for k in range( len(expr_doc['training_doc_lst']) ):

                tf_matrix.append({
                    'screen_name': user['screen_name'],
                    'segments': expr_doc['training_doc_lst'][k]['segments'],
                    'text': combine_text_across_segs_bloc_dims( expr_doc['training_doc_lst'][k]['text'] ),
                    'text_structured': expr_doc['training_doc_lst'][k]['text'],
                    'id': str(i) + ', ' + str(j) + ', ' + str(k),
                    'user_indx': i,
                    'experiment_indx': j,
                    'training_doc_indx': k
                })

                
                logger.debug( '\t\ttraining doc: ' + str(expr_doc['training_doc_lst'][k]) )
                logger.debug( '\t\t user_indx: ' + str(i) )
                logger.debug( '\t\t expr_indx: ' + str(j) )
                logger.debug( '\t\t training_doc_indx: ' + str(k) )
                logger.debug('')
        
        logger.debug('')
        logger.debug('')


    #generate matrix
    if( len(tf_matrix) != 0 ):
        tf_matrix = get_tf_matrix( 
            tf_matrix, 
            n=ngram, 
            token_pattern=token_pattern,
            lowercase=False, 
            add_normalized_tf_matrix=tf_matrix_norm,
            add_tf_idf_matrix=tf_idf_norm,
            rm_doc_text=False,
        )
    
    
    if( matrix_key in tf_matrix ):
        update_user_test_train_tf_vects( tf_matrix[matrix_key], all_screenames_bloc )

    return tf_matrix



def query_user_experiment_sample(user, experiment_type):

    if( 'bloc_experiment_docs' not in user ):
        return []

    experiment_runs = []
    for expr_run in user['bloc_experiment_docs']:

        if( experiment_type == 'testing_doc_lst' ):
            experiment_runs.append( expr_run['testing_doc_lst'] )

    return experiment_runs

def rm_dup_train_screen_name(training_samples):
    
    '''
        Precondition: training segments are already sorted in descending order (see: generate_bloc_expr_template() )

        Remove duplicate training samples, leave just one; the one closest to the test segment
        E.g.,
        Input: (anwala has multiple segment samples, so leave just one, the one closest to the test segment)
        i: 0 shawnmjones, [44] test
            0.35 phonedude_mln train-seg-2 [43, 42] (r ⚁, 0.18)
            0.45 acnwala train-seg-2 [40, 39] (⚁ r, 0.33)
            0.47 acnwala train-seg-2 [41, 40] (⚁ r, 0.32)
            0.48 acnwala train-seg-2 [43, 42] (r ⚁, 0.23)
            0.57 acnwala train-seg-2 [42, 41] (⚂ r, 0.36)

        Output:
        i: 0 shawnmjones, [44] test
            0.35 phonedude_mln train-seg-2 [43, 42] (r ⚁, 0.18)
            0.48 acnwala train-seg-2 [43, 42] (r ⚁, 0.23)
    '''

    new_training_samples = {}
    for sampl in training_samples:
        
        screen_name = sampl['screen_name']
        if( screen_name in new_training_samples ):
            
            #e.g., [43, 42] > [42, 41]
            if( sampl['segments'][0] > new_training_samples[screen_name]['segments'][0] ):
                new_training_samples[screen_name] = sampl
        else:
            new_training_samples[screen_name] = sampl


    dedup_training_samples = []
    for user, train_sampl in new_training_samples.items():
        dedup_training_samples.append( train_sampl )

    return dedup_training_samples

def get_sample_top_ngram(top_per_doc_ngram):
    
    if( len(top_per_doc_ngram) == 0 ):
        top_ngrams = 'NA'
    else:
        top_ngrams = '(' + top_per_doc_ngram[0]['term'] + ', ' + "{0:0.2f}".format(top_per_doc_ngram[0]['term_rate']) + ')'

    return top_ngrams

def get_markov_chains_for_bloc_dims(segment_seqs):

    mkv_inputs = {}
    for segment in segment_seqs:
        for dim, bloc_seq in segment.items():
            if( bloc_seq != '' ):

                bloc_seq = re.sub(r'[ |()*]', '', bloc_seq)
                mkv_inputs.setdefault(dim, [])
                mkv_inputs[dim].append(bloc_seq)


    mkv_chains = {}
    for dim, bloc_seq in mkv_inputs.items():
        mkv_chains[dim] = { 'model': BLOCMarkovChain(training_sequence_lst=bloc_seq), 'training_seq': ''.join(bloc_seq) }

    return mkv_chains

def get_src_tgt_mkv_dist(mc_src, mc_tgt):

    total_src_tgt_prob_all_models = 0
    good_prob_flag = False
    #dist in function name because: smaller log prob is better, if ordinary prob is used (log_prob = False), then this function performs sim est. since larger values are better

    for dim in mc_src:
        if( dim in mc_tgt ):

            good_prob_flag = True
            prob_src_tgt = BLOCMarkovChain.prob_of_sequence( mc_src[dim]['model'].training_model, mc_tgt[dim]['training_seq'], log_prob=True )
            prob_src_tgt = prob_src_tgt/len(mc_tgt[dim]['training_seq'])

            prob_tgt_src = BLOCMarkovChain.prob_of_sequence(mc_tgt[dim]['model'].training_model, mc_src[dim]['training_seq'], log_prob=True)
            prob_tgt_src = prob_tgt_src/len(mc_src[dim]['training_seq'])

            total_src_tgt_prob_all_models += (prob_src_tgt+prob_tgt_src)/2
            
    return total_src_tgt_prob_all_models, good_prob_flag

def pairwise_pred_prob_multi_markvov( tf_matrix ):
    
    X = len(tf_matrix)
    pred_prob_matrix = []

    '''
        ' * ':  separates bloc dimenensions
        ' ** ': separates combined segments
        ' | ':  time segmenter

        For example:
            ⚂T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T * ttttt(Ht)tttttttttttttt ** T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T * tt(Ht)tttt(Ht)t(Ht)(Ht)ttttttttt
            segment 1: ⚂T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T * ttttt(Ht)tttttttttttttt
            segment 2: T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T * tt(Ht)tttt(Ht)t(Ht)(Ht)ttttttttt
    '''
    
    for i in range(X):

        print( '\tpairwise_pred_prob(), user: {} of {}'.format(i, X) )
        
        pred_prob_matrix.append([])
        mc_src = get_markov_chains_for_bloc_dims( tf_matrix[i]['text_structured'] )

        for j in range(X):

            mc_tgt = get_markov_chains_for_bloc_dims( tf_matrix[j]['text_structured'] )
            total_src_tgt_prob_all_models, good_prob_flag = get_src_tgt_mkv_dist( mc_src, mc_tgt )

            if( good_prob_flag ):
                pred_prob_matrix[-1].append( total_src_tgt_prob_all_models )
            else:
                pred_prob_matrix[-1].append( -1000 )

    return pred_prob_matrix

def pairwise_pred_prob( tf_matrix ):
    

    X = len(tf_matrix)
    pred_prob_matrix = []


    print( '\tpairwise_pred_prob(): revise function since BLOCMarkovChain has changed, returning' )
    for i in range(X):
        pred_prob_matrix.append([])
        for j in range(X):
            pred_prob_matrix[-1].append(0.5)
    return pred_prob_matrix

    


    print( '\tpairwise_pred_prob():' )
    '''
        tf_matrix[j]['text']'s structure
        ' * ':  separates bloc dimenensions
        ' ** ': separates combined segments
        ' | ':  time segmenter

        For example:
            ⚂T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T * ttttt(Ht)tttttttttttttt ** T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T * tt(Ht)tttt(Ht)t(Ht)(Ht)ttttttttt
            segment 1: ⚂T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T * ttttt(Ht)tttttttttttttt
            segment 2: T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T⚁T * tt(Ht)tttt(Ht)t(Ht)(Ht)ttttttttt
    '''
    
    for i in range(X):

        print( '\tpairwise_pred_prob(), user: {} of {}'.format(i, X) )
        pred_prob_matrix.append([])
        
        src_seq = re.sub(r'[ |()*]', '', tf_matrix[i]['text'])
        mc_src = BLOCMarkovChain( training_sequence_lst=[src_seq] )

        for j in range(X):

            tgt_seq = re.sub(r'[ |()*]', '', tf_matrix[j]['text'])
            mc_tgt = BLOCMarkovChain( training_sequence_lst=[ tgt_seq ] )
            #prob = mc.prob_of_sequence( tgt_seq, log_prob=True )
            
            prob_src_tgt = BLOCMarkovChain.s_prob_of_sequence(mc_src.training_model, tgt_seq, log_prob=True)
            prob_src_tgt = prob_src_tgt/len(tgt_seq)

            prob_tgt_src = BLOCMarkovChain.s_prob_of_sequence(mc_tgt.training_model, src_seq, log_prob=True)
            prob_tgt_src = prob_tgt_src/len(src_seq)
            
            pred_prob_matrix[-1].append( (prob_src_tgt+prob_tgt_src)/2 )
            #pred_prob_matrix[-1].append( 0 )

    return pred_prob_matrix

def analyze_ngram_vectors( tf_matrix, matrix_key='tf_matrix_normalized', analyze_training_segment_length=1 ):

    if( matrix_key not in tf_matrix ):
        return {
            'experiment_parameters': {
                'matrix_key': matrix_key
            },
            'prediction_results': []
        }

    if( len(tf_matrix[matrix_key]) == 0 ):
        return {
            'experiment_parameters': {
                'matrix_key': matrix_key
            },
            'prediction_results': []
        }
    
    bloc_prediction_results = []
    markov_prediction_results = []

    X = np.array([ user['tf_vector'] for user in tf_matrix[matrix_key] ])

    distance_matrix = pairwise_distances( X, metric='euclidean' )#site 1
    prob_seq_matrix = pairwise_pred_prob( tf_matrix[matrix_key] )

    logger.debug( '\nanalyze_ngram_vectors() for matrix_key:' + matrix_key )
    logger.debug( 'matrix.len: ' + str(len(tf_matrix[matrix_key])) )
    logger.debug( 'dist mat:' + str(distance_matrix.shape) + '\n' )

    for i in range( distance_matrix.shape[0] ):
        
        a = tf_matrix[matrix_key][i]['screen_name'] + ', ' + str(tf_matrix[matrix_key][i]['segments'])
        if( 'training_doc_indx' in tf_matrix[matrix_key][i] ):
            continue

        logger.debug('i: ' + str(i) + ' ' + a + ' test ' + get_sample_top_ngram( tf_matrix['top_ngrams']['per_doc'][i]['ngrams'] ) )

        training_samples = []
        train_dedup_set = set()
        pred_payload = {
            'test': {
                'screen_name': tf_matrix[matrix_key][i]['screen_name'],
                'segments': tf_matrix[matrix_key][i]['segments'],
                'train_non_test_screen_name_count': set()
            }
        }
        
        for j in range( distance_matrix.shape[1] ):

            if( i == j ):
                continue

            train_tf_mat = tf_matrix[matrix_key][j]
            training_ky = train_tf_mat['screen_name'] + ', ' + str(train_tf_mat['segments'])


            if( len(train_tf_mat['segments']) != analyze_training_segment_length ):
                #mismatch of count
                continue

            if( training_ky in train_dedup_set ):
                #this is a duplicate training sample
                continue
            train_dedup_set.add(training_ky)


            if( tf_matrix[matrix_key][i]['screen_name'] == train_tf_mat['screen_name'] ):

                test_set = set(tf_matrix[matrix_key][i]['segments'])
                train_set = set(train_tf_mat['segments'])  
                
                if( len(test_set & train_set) != 0 ):
                    #skip because train and test have segment in common
                    continue

            if( train_tf_mat['segments'][0] >= tf_matrix[matrix_key][i]['segments'][0] ):
                #skip future (e.g, 44) trained to predict current 43
                continue


            top_per_doc_ngram = 'NA'
            if( train_tf_mat['id'] == tf_matrix['top_ngrams']['per_doc'][j]['id'] ):
                top_per_doc_ngram = get_sample_top_ngram( tf_matrix['top_ngrams']['per_doc'][j]['ngrams'] )


            if( tf_matrix[matrix_key][i]['screen_name'] != train_tf_mat['screen_name'] ):
                pred_payload['test']['train_non_test_screen_name_count'].add( train_tf_mat['screen_name'] )

            training_samples.append({
                'screen_name': train_tf_mat['screen_name'],
                'segments': train_tf_mat['segments'],
                'distance': distance_matrix[i][j], #site 2
                'pred_prob': prob_seq_matrix[i][j],
                'label': 'train-seg-' + str( len(train_tf_mat['segments']) ),
                'top_per_doc_ngram': top_per_doc_ngram,
                'text': train_tf_mat['text']
            })

        training_samples = rm_dup_train_screen_name( training_samples )
        bloc_training_samples = sorted( training_samples, key=lambda i: i['distance'] )
        markov_training_samples = sorted( training_samples, key=lambda i: i['pred_prob'], reverse=True )
        
        pred_payload['test']['train_non_test_screen_name_count'] = list(pred_payload['test']['train_non_test_screen_name_count'])
        bloc_prediction_results.append({
            'test': pred_payload['test'],
            'train': bloc_training_samples
        })

        markov_prediction_results.append({
            'test': pred_payload['test'],
            'train': markov_training_samples
        })

        for j in range( len(bloc_training_samples) ):
            train = bloc_training_samples[j]
            logger.debug( '\tbloc ' + "{0:0.2f}".format(train['distance']) + ' ' + train['screen_name'] + ' ' + train['label'] + ' ' + str(train['segments']) + ' ' + train['top_per_doc_ngram'] )
        
        logger.debug('')
        logger.debug('i: ' + str(i) + ' ' + a + ' test ' + tf_matrix[matrix_key][i]['text'] )
        
        for j in range( len(markov_training_samples) ):
            train = markov_training_samples[j]
            logger.debug( '\tmkv ' + "{0:0.2f}".format(train['pred_prob']) + ' ' + train['screen_name'] + ' ' + train['label'] + ' ' + str(train['segments']) + ' ' + train['text'] )
        
        logger.debug('')

    return {
        'experiment_parameters': {
            'matrix_key': matrix_key
        },
        'prediction_results': {
            'bloc': bloc_prediction_results,
            'markov': markov_prediction_results#site 3
        }
    }

def report_prediction_results( pred_results, parameters ):

    retrieved_docs = 0
    relevant_docs = 0
    random_relevant_docs = 0

    for test_train in pred_results:

        if( 'test' not in test_train  or 'train' not in test_train ):
            continue

        if( len(test_train['test']['train_non_test_screen_name_count']) == 0 or len(test_train['train']) == 0 ):
            continue

        retrieved_docs += 1
        
        if( test_train['test']['screen_name'] == test_train['train'][0]['screen_name'] ):
            relevant_docs += 1

        guess = randint( 0, len(test_train['train'])-1 )
        if( test_train['test']['screen_name'] == test_train['train'][guess]['screen_name'] ):
            random_relevant_docs += 1
    
    if( retrieved_docs == 0 ):
        precision = -1
        random_precision = -1
    else:
        precision = relevant_docs/retrieved_docs
        random_precision = random_relevant_docs/retrieved_docs

    return {
        'precision': precision,
        'random_precision': random_precision
    }

def update_user_test_train_tf_vects(tf_matrix, all_screenames_bloc):
    
    if( len(all_screenames_bloc) == 0 ):
        return

    for mat in tf_matrix:

        i = mat['user_indx']
        j = mat['experiment_indx']
        user = all_screenames_bloc[i]

        if( 'training_doc_indx' in mat ):            
            k = mat['training_doc_indx']
            user['bloc_experiment_docs'][j]['training_doc_lst'][k]['tf_vector'] = mat['tf_vector']

        else:
            user['bloc_experiment_docs'][j]['testing_doc_lst']['tf_vector'] = mat['tf_vector']

def analyze_bloc_for_users(all_screenames_bloc, ngram=1, analyze_bloc_dimensions='action', analyze_matrix_key='tf_matrix_normalized', analyze_training_segment_length=1, max_pages=-1, timeline_startdate='', timeline_scroll_by_hours=None, **kwargs):

    kwargs.setdefault('analyze_tf_matrix_norm', 'l1')
    kwargs.setdefault('analyze_tf_idf_matrix_norm', 'l2')
    kwargs.setdefault('token_pattern', r'[^ |()*]')
    kwargs.setdefault('bloc_variant', None)

    new_tf_matrix = {}
    all_experiments = generate_bloc_expr_template( all_screenames_bloc )
    

    gen_expr_doc_lst_for_users( all_screenames_bloc, deepcopy(all_experiments), analyze_bloc_dimensions )
    tf_matrix = gen_ngram_tf_matrix( all_screenames_bloc, ngram, token_pattern=kwargs['token_pattern'], matrix_key=analyze_matrix_key, tf_matrix_norm=kwargs['analyze_tf_matrix_norm'], tf_idf_norm=kwargs['analyze_tf_idf_matrix_norm'] )

    if( kwargs['bloc_variant'] is not None ):

        bloc_variant_tf_matrix = gen_bloc_variant_tf_mat( tf_matrix, kwargs['bloc_variant'] )
        new_tf_matrix = merge_bloc_matrices( tf_matrix, bloc_variant_tf_matrix, kwargs['bloc_variant'] )

        #dumpJsonToFile('tf_matrix.json', tf_matrix)

        tf_matrix = update_bloc_model(tf_matrix, new_tf_matrix, tf_matrix_norm=kwargs['analyze_tf_matrix_norm'], tf_idf_norm=kwargs['analyze_tf_idf_matrix_norm'] )
        if( analyze_matrix_key in tf_matrix ):
            update_user_test_train_tf_vects(tf_matrix[analyze_matrix_key], all_screenames_bloc)

        #dumpJsonToFile('new_tf_matrix.json', tf_matrix)

    pred_res = analyze_ngram_vectors( tf_matrix, matrix_key=analyze_matrix_key, analyze_training_segment_length=analyze_training_segment_length )
    
    pred_res['experiment_parameters']['ngram'] = ngram
    pred_res['experiment_parameters']['analyze_training_segment_length'] = analyze_training_segment_length
    pred_res['experiment_parameters']['bloc_experiment_dimensions'] = analyze_bloc_dimensions

    pred_res['experiment_parameters']['max_pages'] = max_pages
    pred_res['experiment_parameters']['timeline_startdate'] = timeline_startdate
    pred_res['experiment_parameters']['timeline_scroll_by_hours'] = timeline_scroll_by_hours

    pred_report_bloc = { 'precision': -1, 'random_precision': -1 }
    pred_report_markov = { 'precision': -1, 'random_precision': -1 }

    if( len(pred_res['prediction_results']) != 0 ):
        pred_report_bloc = report_prediction_results( pred_res['prediction_results']['bloc'], pred_res['experiment_parameters'] )
        pred_report_markov = report_prediction_results( pred_res['prediction_results']['markov'], pred_res['experiment_parameters'] )

    return {
        'all_screenames_bloc': all_screenames_bloc, #conforming to how bloc.proc_req() returns 
        'prediction_report': {'bloc': pred_report_bloc, 'markov': pred_report_markov},
        'experiment_parameters': pred_res['experiment_parameters']
    }


#!/usr/bin/env python

from setuptools import setup, find_packages

desc = """Behavioral Language for Online Classification (BLOC): A General Language for Modeling Social Media Account Behavior."""

__appversion__ = None

#__appversion__, defined here
exec(open('bloc/version.py').read())


setup(
    name='twitterbloc',
    version=__appversion__,
    description=desc,
    long_description='See: https://github.com/anwala/bloc',
    author='Alexander C. Nwala',
    author_email='acnwala@wm.edu',
    url='https://github.com/anwala/bloc',
    packages=find_packages(),
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    package_data={
        'bloc': [
            './symbols.json'
        ]
    },
    install_requires=[
        'numpy',
        'osometweet',
        'requests-oauthlib',
        'sklearn',
        'textblob'
    ],
    scripts=[
        'bin/bloc'
    ]
)



import json
import logging
import numpy as np
import warnings

from copy import deepcopy
from datetime import datetime
from sklearn.feature_extraction.text import CountVectorizer

from bloc.util import genericErrorInfo
from bloc.util import get_color_txt
from bloc.util import getDictFromJsonGZ
from bloc.util import gzipTextFile

logger = logging.getLogger('bloc.bloc')

class BLOCMarkovChain:
    
    '''
        Class for trianing a BLOC Markov Model. See use_cases() for usage examples
    '''
    def __init__(self, training_sequence_lst=None, model=None, model_output_filename=None, model_label='', states=None, init_prob_dist=None, transition_matrix=None, **kwargs):
            
        if( training_sequence_lst is not None ):
            #create model from training data
            training_model = BLOCMarkovChain.train_markov_model(training_sequence_lst, model_output_filename, model_label)

        elif( model is not None ):
            #instantiate markov chain from model
            training_model = model
        else:
            training_model = None

        
        self.states = states
        self.transition_matrix = transition_matrix
        self.init_prob_dist = init_prob_dist
        self.training_model = training_model
        self.model_label = model_label

        BLOCMarkovChain.validate_prob_params(self.states, self.transition_matrix, self.init_prob_dist)
    
    @staticmethod
    def validate_prob_params(states, transition_matrix, init_prob_dist):

        valid_flag = True


        if( states is None ):
            valid_flag = False
        elif( len(states) == 0 ):
            warnings.warn('States is empty.')
            valid_flag = False


        if( transition_matrix is None ):
            valid_flag = False
        else:
            transition_matrix = np.array(transition_matrix)
            if not np.all(transition_matrix >= 0):
                warnings.warn('Transition matrix contains negative value(s).')
                valid_flag = False
            
            for i in range( transition_matrix.shape[0] ):
                sm = np.sum( transition_matrix[i] )
                if not np.isclose( sm, 1. ):
                    warnings.warn( 'Transition probabilities don\'t sum to 1. Row {}, Sum = {}'.format(i, sm) )
                    valid_flag = False


        if( init_prob_dist is None ):
            valid_flag = False
        elif not np.isclose(np.sum(init_prob_dist), 1.):
            warnings.warn('Initial probabilities distribution don\'t sum to 1.')
            valid_flag = False

        return valid_flag

    @staticmethod
    def train_markov_model(sequence_lst, model_output_filename=None, model_label=''):
        
        ngram_freq = BLOCMarkovChain.get_ngram_freq(sequence_lst)
        mk_freq_mat = BLOCMarkovChain.get_markov_freq_matrix(ngram_freq)

        if( model_output_filename is not None ):
            if( 'markov_freq_matrix' in mk_freq_mat ):

                mk_freq_mat['markov_freq_matrix'] = list( mk_freq_mat['markov_freq_matrix'] )

                for i in range( len(mk_freq_mat['markov_freq_matrix']) ):
                    mk_freq_mat['markov_freq_matrix'][i] = [ float(a) for a in mk_freq_mat['markov_freq_matrix'][i] ]

                mk_freq_mat['model_label'] = model_label
                mk_freq_mat['model_created_at_utc'] = datetime.utcnow().isoformat().split('.')[0] + 'Z'
                gzipTextFile(model_output_filename, json.dumps(mk_freq_mat, ensure_ascii=False))

        return mk_freq_mat

    @staticmethod
    def mle_get_markov_params( markov_freq_mat, unknown_seq ):

        if( 'vocab' not in markov_freq_mat or 'markov_freq_matrix' not in markov_freq_mat or unknown_seq == '' ):
            logger.info("\tmle_get_markov_params(): 'vocab' not in markov_freq_mat or 'markov_freq_matrix' not in markov_freq_mat or unknown_seq == ''")
            return markov_freq_mat

        if( len(markov_freq_mat['vocab']) == 0 ):
            logger.info("\tmle_get_markov_params(): len(markov_freq_mat['vocab']) == 0")
            return markov_freq_mat


        if( isinstance(markov_freq_mat['markov_freq_matrix'], list) ):
            markov_freq_mat['markov_freq_matrix'] = np.array( markov_freq_mat['markov_freq_matrix'] )

        
        start_prob_dist = [ markov_freq_mat['start_dist'][s] for s in markov_freq_mat['vocab'] ]

        #find/add unseen states - start
        new_states = []
        for s in set(unknown_seq):
            if( s not in markov_freq_mat['vocab'] ):
                new_states.append(s)
                markov_freq_mat['start_dist'][s] = 0

        
        new_states_count = len(new_states)
        markov_freq_mat['vocab'] += new_states

        row, col = markov_freq_mat['markov_freq_matrix'].shape
        z = np.zeros(( row, new_states_count ))
        
        #add extra columns of zeros for new states
        markov_freq_mat['markov_freq_matrix'] = np.concatenate( (markov_freq_mat['markov_freq_matrix'], z), axis=1 )
        
        #add extra rows of zeros to new matrix
        z = np.zeros(( new_states_count, len(markov_freq_mat['vocab']) ))
        markov_freq_mat['markov_freq_matrix'] = np.concatenate( (markov_freq_mat['markov_freq_matrix'], z), axis=0 )
        #find/add unseen states - end


        #gen (smoothed) start_dist - start
        start_prob_dist += [0] * new_states_count
        if( 0 in start_prob_dist ):
            start_prob_dist = [ v+1 for v in start_prob_dist ]

        total_freq = sum(start_prob_dist)
        markov_freq_mat['start_prob_dist'] = [ v/total_freq for v in start_prob_dist ]
        #gen (smoothed) start_dist - end

        
        #calc transition probs, smooth if necessary 
        markov_freq_matrix_row_sums = [ sum(v) for v in markov_freq_mat['markov_freq_matrix'] ]
        #new_states        


        n = len(markov_freq_mat['vocab'])
        markov_freq_mat['markov_trans_prob_matrix'] = np.zeros( (n,n) )

        '''
            print('vocab/transition_matrix:')
            print(markov_freq_mat['vocab'])
            print(markov_freq_mat['markov_freq_matrix'])
            print('all transition prob')
        '''
        if( 'un_smoothed_markov_freq_matrix' not in markov_freq_mat ):
            markov_freq_mat['un_smoothed_markov_freq_matrix'] = deepcopy( markov_freq_mat['markov_freq_matrix'] )

        for i in range(n):

            s0 = markov_freq_mat['vocab'][i]

            if( 0 in markov_freq_mat['markov_freq_matrix'][i] ):
                markov_freq_mat['markov_freq_matrix'][i] = markov_freq_mat['markov_freq_matrix'][i] + 1
                markov_freq_matrix_row_sums[i] = sum( markov_freq_mat['markov_freq_matrix'][i] )

            sm = markov_freq_matrix_row_sums[i]

            for j in range(n):

                s1 = markov_freq_mat['vocab'][j]
                ky = s0 + ' ' + s1

                prob = markov_freq_mat['markov_freq_matrix'][i, j]/sm
                markov_freq_mat['markov_trans_prob_matrix'][i, j] = prob
                
                #print( '{}, {}, sq: {}, C({}|{}) = {}, C(*|{}) = {}'.format(i, j, ky, s1, s0, markov_freq_mat['markov_freq_matrix'][i, j], s0, sm), 'P({}|{}) = {:.2f}'.format(s1, s0, prob) )

        #print('markov_trans_prob_matrix')
        #print( markov_freq_mat['markov_trans_prob_matrix'] )
        return markov_freq_mat

    @staticmethod
    def get_markov_freq_matrix(ngram_freq):
        
        payload = {
            'vocab': [],
            'start_dist': ngram_freq.get('start_dist', {}),
            'un_smoothed_markov_freq_matrix': np.array([]),
            'markov_freq_matrix': np.array([])
        }

        if( 'vocab' not in ngram_freq or 'ngram_freq' not in ngram_freq ):
            return payload

        if( 'all_docs' not in ngram_freq['ngram_freq'] ):
            return payload


        #generate markov chain states
        markov_states = []
        dedup_set = set()
        for v in ngram_freq['vocab']:
            
            v_split = v.split(' ')

            if( len(v_split) != 1 or v in dedup_set ):
                continue

            dedup_set.add(v)
            markov_states.append(v)
        
        n = len(markov_states)
        markov_freq_matrix = np.zeros((n, n))


        #prerequisite for markov chain transition matrix: populate markov_freq_matrix (pre-cursor to markov_transition_matrix) with counts that would be used to calc P(x|y)
        for i in range(n):

            s0 = markov_states[i]
            for j in range(n):
                
                s1 = markov_states[j]
                ky = s0 + ' ' + s1

                if( ky not in ngram_freq['ngram_freq']['all_docs'] ):
                    continue

                markov_freq_matrix[i, j] = ngram_freq['ngram_freq']['all_docs'][ky]

            #set missing start_dist - start
            payload['start_dist'].setdefault(s0, 0)
            #set missing start_dist - end
    
        
        #for debug - start
        '''
        print('\nmarkov_states', markov_states)
        for i in range(n):

            s0 = markov_states[i]
            for j in range(n):

                s1 = markov_states[j]
                ky = s0 + ' ' + s1
                sm = sum(markov_freq_matrix[i])

                print( 'ky: {}, C({}|{}) = {}, C(*|{}) = {}'.format(ky, s1, s0, markov_freq_matrix[i, j], s0, sm) )

        print('\nmarkov_freq_matrix:')
        print(markov_freq_matrix)
        '''
        #for debug - end

        payload['vocab'] = markov_states
        payload['markov_freq_matrix'] = markov_freq_matrix
        payload['un_smoothed_markov_freq_matrix'] = deepcopy(markov_freq_matrix)
        return payload

    @staticmethod
    def get_ngram_freq(sequence_lst, ngram_range=(1, 2), token_pattern=r'[^ |()*]', **kwargs):
        
        kwargs.setdefault('lowercase', False)
        payload = {
            'vocab': [],
            'start_dist': {},
            'ngram_freq': {
                'all_docs': {},
                'all_docs_total_tf': {}
            }
        }


        for seq in sequence_lst:
            
            if( seq == '' ):
                continue

            payload['start_dist'].setdefault( seq[0], 0 )
            payload['start_dist'][ seq[0] ] += 1


        try:
            count_vectorizer = CountVectorizer(
                stop_words=None, 
                token_pattern=token_pattern, 
                ngram_range=ngram_range,
                lowercase=kwargs['lowercase']
            )

            payload['tf_matrix'] = count_vectorizer.fit_transform(sequence_lst).toarray()
            payload['vocab'] = count_vectorizer.get_feature_names()

            #convert types for JSON serialization - start
            for opt in ['tf_matrix']:

                if( opt not in payload ):
                    continue

                payload[opt] = list( payload[opt] )
                for i in range( len(payload[opt]) ):
                    payload[opt][i] = [ float(a) for a in payload[opt][i] ]
            #convert types for JSON serialization - end
        except:
            genericErrorInfo()
            return payload


        for i in range( len(payload['tf_matrix']) ):

            total_tf = sum(payload['tf_matrix'][i])
            if( total_tf == 0 ):
                total_tf = -1

            single_doc_tf = [ {'term': v, 'term_freq': int(payload['tf_matrix'][i][j]), 'term_rate': payload['tf_matrix'][i][j]/total_tf} for (j, v) in enumerate(payload['vocab']) if payload['tf_matrix'][i][j] != 0 ]
            
            for tf_dct in single_doc_tf:
                payload['ngram_freq']['all_docs'].setdefault( tf_dct['term'], 0 )
                payload['ngram_freq']['all_docs'][ tf_dct['term'] ] += int(tf_dct['term_freq'])

                term_ngram = '{}-gram'.format( len(tf_dct['term'].split(' ')) )
                
                payload['ngram_freq']['all_docs_total_tf'].setdefault( term_ngram, 0 )
                payload['ngram_freq']['all_docs_total_tf'][ term_ngram ] += int(tf_dct['term_freq'])    

        del payload['tf_matrix']
        return payload

    def frm_prob_of_seq_mle_get_markov_params(self, sequence):

        #here means object was created without instantiating states or init_prob_dist or transition_matrix
        if( self.training_model is None ):
            return None
        else:
            #Attempt to instantiate states or init_prob_dist or transition_matrix from training_model
            markov_params = BLOCMarkovChain.mle_get_markov_params(self.training_model, sequence)
            
            self.states = markov_params.get('vocab', None)
            self.init_prob_dist = markov_params.get('start_prob_dist', None)
            self.transition_matrix = markov_params.get('markov_trans_prob_matrix', None)

            if( BLOCMarkovChain.validate_prob_params(self.states, self.transition_matrix, self.init_prob_dist) is False ):
                return None

            return True

    def prob_of_sequence(self, sequence, log_prob=True):
        
        #print('prob_of_sequence({}):'.format(sequence))
        if( self.states is None or self.init_prob_dist is None or self.transition_matrix is None ):
            #print('\t0, None: vocab: {}, init_prob_dist: {}, transition_matrix: {}'.format(self.states is None, self.init_prob_dist is None, self.transition_matrix is None))            
            
            #here means object was created without instantiating states or init_prob_dist or transition_matrix
            if( self.frm_prob_of_seq_mle_get_markov_params(sequence) is None ):
                return None
        else:
            #print('\t1')
            seq_min_states = set(sequence) - set(self.states)
            #print('\tseq - states:', seq_min_states)
            if( len(seq_min_states) != 0 ):
                #here means that the sequence has an unseen state, so retrain model and smooth for unseen state
                if( self.frm_prob_of_seq_mle_get_markov_params(sequence) is None ):
                    return None

            


        if( isinstance(sequence, str) ):
            sequence = list(sequence)

        start = sequence[0]
        prob = self.states.index( start )
        prob = self.init_prob_dist[ prob ]

        if( log_prob is True ):
            prob = np.log(prob)
        
        #logger.info( '\nprob_of_sequence():' )
        #logger.info( '\tstates: {}'.format(self.states) )
        #logger.info( '\tP({}):'.format(sequence) )
        #logger.info( '\n\tP({}) = {:4f}'.format(start, prob) )
        for i in range( 1, len(sequence) ):
            
            cur_indx = self.states.index( sequence[i-1] )
            next_indx = self.states.index( sequence[i] )
            trans_prob = self.transition_matrix[cur_indx, next_indx]
            na = 1 if self.training_model['un_smoothed_markov_freq_matrix'][cur_indx, next_indx] == 0 else 0
            
            if( log_prob is True ):
                trans_prob = np.log( trans_prob )
                prob += trans_prob
            else:
                prob *= trans_prob
            
            #logger.info( '\tP({}|{}) = {:.4f} (NA: {})'.format(sequence[i], sequence[i-1], trans_prob, na) )

        return prob

    @staticmethod
    def s_prob_of_sequence(training_model, sequence, log_prob=True):
                
        #Attempt to instantiate states or init_prob_dist or transition_matrix from training_model
        markov_params = BLOCMarkovChain.mle_get_markov_params(deepcopy(training_model), sequence)
        
        states = markov_params.get('vocab', None)
        init_prob_dist = markov_params.get('start_prob_dist', None)
        transition_matrix = markov_params.get('markov_trans_prob_matrix', None)
        un_smoothed_markov_freq_matrix = markov_params.get('un_smoothed_markov_freq_matrix', None)

        #print('\t\tv:', states)
        #print('\t\ts:', markov_params.get('start_dist', None))
        #if( BLOCMarkovChain.validate_prob_params(states, transition_matrix, init_prob_dist) is False ):
        #    return None
        #print('vocab/states:', states)
        #print('markov_freq_matrix:\n', markov_params['markov_freq_matrix'])
        #print('transition_matrix:\n', transition_matrix)


        if( isinstance(sequence, str) ):
            sequence = list(sequence)

        start = sequence[0]
        prob = states.index( start )
        prob = init_prob_dist[ prob ]
        #prob = [1/len(states)]*len(states)

        if( log_prob is True ):
            prob = np.log(prob)
        
        
        #logger.info( '\nprob_of_sequence():' )
        #logger.info( '\tstates: {}'.format(states) )
        #logger.info( '\tP({}):'.format(sequence) )
        #logger.info( '\n\tP({}) = {:4f}'.format(start, prob) )
        
        #print( '\nprob_of_sequence():' )
        #print( '\tstates: {}'.format(states) )
        #print( '\tP({}):'.format(sequence) )
        #print( '\n\tP({}) = {:4f}'.format(start, prob) )
        #print('un_smoothed_markov_freq_matrix:\n', un_smoothed_markov_freq_matrix)
        #print('\ntransition_matrix:')
        #print(transition_matrix)

        for i in range( 1, len(sequence) ):
            
            cur_indx = states.index( sequence[i-1] )
            next_indx = states.index( sequence[i] )
            trans_prob = transition_matrix[cur_indx, next_indx]

            #print('\tcur_indx:', cur_indx)
            #print('\tnext_indx:', next_indx)
            #print()
            #na = 1 if un_smoothed_markov_freq_matrix[cur_indx, next_indx] == 0 else 0

            if( log_prob is True ):
                trans_prob = np.log( trans_prob )
                prob += trans_prob
            else:
                prob *= trans_prob
            
            #logger.info( '\tP({}|{}) = {:.4f} (NA: {})'.format(sequence[i], sequence[i-1], trans_prob, na) )

        return prob

    @staticmethod
    def matrix_to_str( matrix, col_mat, vocab, model_label, mat_name ):
        
        if( len(matrix) == 0 or len(vocab) == 0 or matrix.shape != col_mat.shape ):
            return ''

        N = len(matrix) 
        all_prnt = ''
        dec = 5#match with: val = '{:5}'.format(val)

        heading = (' '*dec).join(['{}'.format(v) for v in vocab])
        all_prnt += model_label + f'{mat_name}:\n' + ' '*(dec*2) + heading + '\n'
        
        for i in range(N):
            
            prnt = ''
            row = matrix[i]
            for j in range( len(row) ):
                val = row[j]
                val = '{:5}'.format(val)
                
                if( col_mat[i, j] == 0 ):
                    prnt += '{} '.format(val)
                else:
                    prnt += '{} '.format( getColorTxt(val) )

            all_prnt += ' '*dec + vocab[i] + ' ' + prnt
            
            if( i < N-1 ):
                all_prnt += '\n'

        return all_prnt

    def __str__(self):
        
        model_label = self.model_label + '\n' if self.model_label != '' else ''

        to_prnt_frq_mat = BLOCMarkovChain.matrix_to_str( 
            self.training_model['markov_freq_matrix'], 
            self.training_model['un_smoothed_markov_freq_matrix'],
            self.training_model['vocab'], 
            model_label, 
            'markov_freq_matrix'
        )

        return to_prnt_frq_mat

